"""OSO Energy constants."""
from typing import Any

# pylint: disable=line-too-long
# HTTP return codes.
HTTP_OK = 200
HTTP_CREATED = 201
HTTP_ACCEPTED = 202
HTTP_MOVED_PERMANENTLY = 301
HTTP_BAD_REQUEST = 400
HTTP_UNAUTHORIZED = 401
HTTP_FORBIDDEN = 403
HTTP_NOT_FOUND = 404
HTTP_METHOD_NOT_ALLOWED = 405
HTTP_UNPROCESSABLE_ENTITY = 422
HTTP_TOO_MANY_REQUESTS = 429
HTTP_INTERNAL_SERVER_ERROR = 500
HTTP_BAD_GATEWAY = 502
HTTP_SERVICE_UNAVAILABLE = 503

OSOTOHA = {
    "Hotwater": {
        "HeaterState": {"on": "on", "off": "off"},
        "HeaterStateBool": {"on": True, "off": False},
        "DeviceConstants": {"minTemp": 10, "maxTemp": 80},
        "HeaterConnection": {None: False, "Connected": True},
        "HeaterMode": {
            None: "off",
            "auto": "auto",
            "manual": "manual",
            "off": "off",
            "Legionella": "legionella",
            "PowerSave": "powerSave",
            "ExtraEnergy": "extraEnergy",
            "Voltage": "voltage"
        },
        "HeaterOptimizationMode": {None: "off"},
        "HeaterSubOptimizationMode": {None: None},
        "HeaterPowerSaveMode": {None: "off", False: "off", True: "on"},
        "HeaterPowerSaveModeBool": {None: False, False: False, True: True},
        "HeaterExtraEnergyMode": {None: "off", False: "off", True: "on"},
        "HeaterExtraEnergyModeBool": {None: False, False: False, True: True},
    }
}

binary_sensor_commands = {
    "POWER_SAVE": "self.session.attr.get_power_save_bool(device.device_id)",
    "EXTRA_ENERGY": "self.session.attr.get_extra_energy_bool(device.device_id)",
    "HEATER_STATE": "self.session.attr.get_heater_state_bool(device.device_id)",
}

sensor_commands = {
    "POWER_LOAD": "self.session.attr.get_actual_load_kwh(device.device_id)",
    "VOLUME": "self.session.attr.get_volume(device.device_id)",
    "TAPPING_CAPACITY": "self.session.attr.get_tapping_capacity(device.device_id)",
    "CAPACITY_MIXED_WATER_40": "self.session.attr.get_capacity_mixed_water_40(device.device_id)",
    "HEATER_MODE": "self.session.attr.get_heater_mode(device.device_id)",
    "OPTIMIZATION_MODE": "self.session.attr.get_optimization_mode(device.device_id)",
    "V40_MIN": "self.session.attr.get_v40_min(device.device_id)",
    "V40_LEVEL_MIN": "self.session.attr.get_v40_level_min(device.device_id)",
    "V40_LEVEL_MAX": "self.session.attr.get_v40_level_max(device.device_id)",
    "PROFILE": "self.session.attr.get_profile(device.device_id)",
    "TEMPERATURE_ONE": "self.session.attr.get_temperature_one(device.device_id)",
    "TEMPERATURE_LOW": "self.session.attr.get_temperature_low(device.device_id)",
    "TEMPERATURE_MID": "self.session.attr.get_temperature_mid(device.device_id)",
    "TEMPERATURE_TOP": "self.session.attr.get_temperature_top(device.device_id)",
}

switch_commands = {
    "HOLIDAY_MODE": "self.session.attr.get_power_save_bool(device.device_id)",
}


class OSOEnergyEntityBase:
    device_id: str
    device_type:str
    device_name: str
    ha_name: str
    ha_type: str
    available: bool
    online: bool

class OSOEnergyWaterHeaterData(OSOEnergyEntityBase):
    """Water heater object containing the device data"""
    current_operation: str
    optimization_mode: str
    heater_state: str
    heater_mode: str
    current_temperature: float
    target_temperature: float
    target_temperature_high: float
    target_temperature_low: float
    min_temperature: float
    max_temperature: float
    profile: list[float]
    power_load: float
    volume: float
    isInPowerSave: bool

class OSOEnergySensorData(OSOEnergyEntityBase):
    """Sensor object containing the device data"""
    osoEnergyType: str
    state: Any

class OSOEnergyBinarySensorData(OSOEnergyEntityBase):
    """Sensor object containing the device data"""
    osoEnergyType: str
    state: bool

class OSOEnergySwitchData(OSOEnergyEntityBase):
    """Switch object containing the device data"""
    osoEnergyType: str
    state: bool



"""OSO Energy API Module."""

import operator
from typing import Optional
from numpy import number

import urllib3
from aiohttp import ClientResponse, ClientSession
from aiohttp.web_exceptions import HTTPError

from ..helper.const import HTTP_UNAUTHORIZED, HTTP_FORBIDDEN
from ..helper.osoenergy_exceptions import NoSubscriptionKey

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class OSOEnergyApiAsync:
    """OSO Energy API Code."""

    def __init__(
            self,
            osoenergy_session=None,
            websession: Optional[ClientSession] = None):
        """Init the api."""
        self.base_url = "https://api.osoenergy.no/water-heater-api"
        self.urls = {
            "devices": self.base_url + "/1/Device/All",
            "turn_on": self.base_url + "/1/Device/{0}/TurnOn?fullUtilizationParam={1}",
            "turn_off": self.base_url + "/1/Device/{0}/TurnOff?fullUtilizationParam={1}",
            "profile": self.base_url + "/1/Device/{0}/Profile",
            "optimization_mode": self.base_url + "/1/Device/{0}/OptimizationMode",
            "set_v40_min": self.base_url + "/1/Device/{0}/V40Min/{1}",
            "enable_holiday_mode": self.base_url + "/1/Device/{0}/HolidayMode/{1}/{2}",
            "disable_holiday_mode": self.base_url + "/1/Device/{0}/HolidayMode",
            "user": self.base_url + "/1/User/Details"
        }
        self.headers = {
            "content-type": "application/json",
            "Accept": "*/*"
        }
        self.timeout = 10
        self.json_return = {
            "original": "No response to OSO Energy API request",
            "parsed": "No response to OSO Energy API request",
        }
        self.session = osoenergy_session
        self.websession = ClientSession() if websession is None else websession

    async def request(self, method: str, url: str, **kwargs) -> ClientResponse:
        """Make a request."""
        data = kwargs.get("data", None)

        if not self.session.subscription_key:
            raise NoSubscriptionKey

        self.headers.update(
            {"Ocp-Apim-Subscription-Key": self.session.subscription_key}
        )

        async with self.websession.request(
            method, url, headers=self.headers, data=data
        ) as resp:
            await resp.json(content_type=None)

            self.json_return.update({"original": resp.status})
            self.json_return.update({"parsed": await resp.json(content_type=None)})

        if operator.contains(str(resp.status), "20"):
            return True

        if resp.status == HTTP_UNAUTHORIZED:
            self.session.logger.error(
                f"Subscription key not authorized when calling {url} - "
                f"HTTP status is - {resp.status}"
            )
        elif resp.status == HTTP_FORBIDDEN:
            self.session.logger.error(
                f"Subscription key not authorized when calling {url} - "
                f"HTTP status is - {resp.status}"
            )
        elif url is not None and resp.status is not None:
            self.session.logger.error(
                f"Something has gone wrong calling {url} - "
                f"HTTP status is - {resp.status}"
            )

        return False

    async def get_user_details(self):
        """Get user details."""
        url = self.urls["user"]
        try:
            await self.request("get", url)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return

    async def get_devices(self):
        """Call the get devices endpoint."""
        url = self.urls["devices"]
        try:
            await self.request("get", url)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return

    async def turn_on(self, device_id: str, full_utilization: bool):
        """Call the get V40 Min endpoint."""
        url = self.urls["turn_on"].format(device_id, full_utilization)
        try:
            await self.request("post", url)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return

    async def turn_off(self, device_id: str, full_utilization: bool):
        """Call the get V40 Min endpoint."""
        url = self.urls["turn_off"].format(device_id, full_utilization)
        try:
            await self.request("post", url)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return

    async def set_profile(self, device_id: str, **kwargs):
        """Call the get V40 Min endpoint."""
        jsc = (
            "{"
            + ",".join(
                ('"' + str(i) + '": ' + str(t) for i, t in kwargs.items())
            )
            + "}"
        )

        url = self.urls["profile"].format(device_id)
        try:
            await self.request("put", url, data=jsc)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return

    async def set_optimization_mode(self, device_id: str, **kwargs):
        """Call the get V40 Min endpoint."""
        jsc = (
            "{"
            + ",".join(
                ('"' + str(i) + '": ' + str(t) + ' ' for i, t in kwargs.items())
            )
            + "}"
        )
        url = self.urls["optimization_mode"].format(device_id)
        try:
            await self.request("put", url, data=jsc)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return

    async def set_v40_min(self, device_id: str, v40_min: number):
        """Call the get V40 Min endpoint."""
        url = self.urls["set_v40_min"].format(device_id, v40_min)
        try:
            await self.request("put", url)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return

    async def enable_holiday_mode(self, device_id: str, start_date: str, end_date: str):
        """Enable holiday mode."""
        url = self.urls["enable_holiday_mode"].format(device_id, start_date, end_date)
        try:
            await self.request("post", url)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return
    
    async def disable_holiday_mode(self, device_id: str):
        """Disable holiday mode."""
        url = self.urls["disable_holiday_mode"].format(device_id)
        try:
            await self.request("delete", url)
        except (OSError, RuntimeError, ZeroDivisionError) as exception:
            raise HTTPError from exception

        return self.json_return


"""OSO Energy Switch Module."""

from aiohttp.web_exceptions import HTTPError
from .helper.const import switch_commands, OSOEnergySwitchData
from datetime import datetime, timezone, timedelta

class OSOEnergySwitch:
    """OSO Energy Switch Code.
    
    Returns:
        object: Switch Object
    """

    switchType = "Switch"

    async def enable_holiday_mode(self, device: OSOEnergySwitchData, period_days: int = 365):
        """Enable holiday mode for device.

        Args:
            device (OSOEnergySwitchData): Device to enable holiday mode for.
            period_days (int, optional): Number of days to enable holiday mode for. Defaults to 365.

        Returns:
            boolean: return True/False if enabling holiday mode was successful.
        """
        final = False
        try:
            start_date = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            end_date = (datetime.now(timezone.utc) + timedelta(days=period_days)).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            resp = await self.session.api.enable_holiday_mode(device.device_id, start_date, end_date)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final
    
    async def disable_holiday_mode(self, device: OSOEnergySwitchData):
        """Disable holiday mode for device.

        Args:
            device (OSOEnergySwitchData): Device to disable holiday mode for.

        Returns:
            boolean: return True/False if disabling holiday mode was successful.
        """
        final = False

        try:
            resp = await self.session.api.disable_holiday_mode(device.device_id)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final

class Switch(OSOEnergySwitch):
    """Home Assistant switch code.

    Args:
        OSOEnergySwitch (object): OSO Energy switch code.
    """

    def __init__(self, session: object = None):
        """Initialise switch.

        Args:
            session (object, optional): session to interact with OSO Energy. Defaults no None.
        """
        self.session = session

    async def get_switch(self, device: OSOEnergySwitchData) -> OSOEnergySwitchData:
        # pylint: disable=eval-used
        """Get updated switch data.

        Args:
            device (OSOEnergySwitchData): Device to update.

        Returns:
            OSOEnergySwitchData: Updated device.
        """
        device.online = await self.session.attr.online_offline(device.device_id)
        if device.online:
            self.session.helper.device_recovered(device.device_id)
            dev_data = OSOEnergySwitchData()
            dev_data.online = device.online
            dev_data.ha_name = device.ha_name
            dev_data.ha_type = device.ha_type
            dev_data.osoEnergyType = device.osoEnergyType
            dev_data.device_id = device.device_id
            dev_data.device_type = device.device_type
            dev_data.device_name = device.device_name
            dev_data.available = await self.session.attr.online_offline(device.device_id)

            if dev_data.osoEnergyType in switch_commands:
                code = switch_commands.get(dev_data.osoEnergyType)
                dev_data.state = await eval(code)

            self.session.switches.update({device.device_id: dev_data})
            return self.session.switches[device.device_id]

        await self.session.log.error_check(
            device.device_id, device.online)
        return device


"""OSO Energy Session Module."""
import asyncio
import copy
import operator
import time
import traceback
from datetime import datetime, timedelta

from aiohttp.web import HTTPException
from apyosoenergyapi import API
from apyosoenergyapi.helper.osoenergy_helper import OSOEnergyHelper
from typing import Any

from .device_attributes import OSOEnergyAttributes
from .helper.const import OSOTOHA, OSOEnergyBinarySensorData, OSOEnergySensorData, OSOEnergySwitchData, OSOEnergyWaterHeaterData
from .helper.osoenergy_exceptions import (
    OSOEnergyApiError,
    OSOEnergyReauthRequired,
    OSOEnergyUnknownConfiguration,
)
from .helper.logger import Logger
from .helper.map import Map


class OSOEnergySession:
    # pylint: disable=no-member
    # pylint: disable=too-many-instance-attributes
    """OSO Energy Session Code.

    Raises:
        HTTPException: HTTP error has occured

    Returns:
        object: Session object
    """

    def __init__(
        self, subscription_key: str, websession: object = None

    ):
        """Initialise the base variable values.

        Args:
            subscription_key (str, reqired): OSO Energy user subscription key.
            websession (object, optional): Websession for api calls. Defaults to None.
        """
        self.subscription_key = subscription_key

        self.helper = OSOEnergyHelper(self)
        self.api = API(osoenergy_session=self, websession=websession)
        self.attr = OSOEnergyAttributes(self)
        self.log = Logger(self)
        self.update_lock = asyncio.Lock()
        self.config = Map(
            {
                "error_list": {},
                "file": False,
                "last_updated": datetime.now(),
                "scan_interval": timedelta(seconds=30),
                "sensors": False,
            }
        )
        self.data = Map(
            {
                "devices": {},
            }
        )
        self.devices = {}
        self.sensors = {}
        self.binary_sensors = {}
        self.switches = {}
        self.device_list = {
            "binary_sensor": [],
            "sensor": [],
            "water_heater": [],
            "switch": [],
        }

    async def update_interval(self, new_interval: timedelta):
        """Update the scan interval.

        Args:
            new_interval (int): New interval for polling.
        """
        if isinstance(new_interval, int):
            new_interval = timedelta(seconds=new_interval)

        interval = new_interval
        if interval < timedelta(seconds=15):
            interval = timedelta(seconds=15)
        self.config.scan_interval = interval

    async def update_subscription_key(self, subscription_key: str):
        """Update subscription key.

        Args:
            subscription_key (dict): The user subscription key.

        Returns:
            str: Subscription key
        """
        self.subscription_key = subscription_key

        return subscription_key

    async def update_data(self):
        """Get latest data for OSO Energy - rate limiting.

        Returns:
            boolean: True/False if update was successful
        """
        await self.update_lock.acquire()
        updated = False
        try:
            next_update = self.config.last_update + self.config.scan_interval
            if datetime.now() >= next_update:
                await self.get_devices()
                updated = True
        finally:
            self.update_lock.release()

        return updated

    async def get_user_email(self):
        """Get user email address
        
        Raises:
            HTTPException: HTTP error has occured loading the user email.

        Returns:
            string: The authenticated user email.
        """
        user_email = None
        api_resp_d = None

        try:
            api_resp_d = await self.api.get_user_details()
            if operator.contains(str(api_resp_d["original"]), "20") is False:
                raise HTTPException
            
            user_email = api_resp_d["parsed"].get("email", None)
            if(user_email == "" or user_email is None):
                raise OSOEnergyApiError
        except (OSError, RuntimeError, OSOEnergyApiError, ConnectionError, HTTPException):
            user_email = None
        
        return user_email

    async def get_devices(self):
        """Get latest device list for the user.

        Raises:
            HTTPException: HTTP error has occured updating the devices.

        Returns:
            boolean: True/False if update was successful.
        """
        get_devices_successful = False
        api_resp_d = None

        try:
            api_resp_d = await self.api.get_devices()
            if operator.contains(str(api_resp_d["original"]), "20") is False:
                raise HTTPException

            if api_resp_d["parsed"] is None:
                raise OSOEnergyApiError

            api_resp_p = api_resp_d["parsed"]
            tmp_devices = {}
            for a_device in api_resp_p:
                tmp_devices.update({a_device["deviceId"]: a_device})

            if len(tmp_devices) > 0:
                self.data.devices = copy.deepcopy(tmp_devices)

            self.config.last_update = datetime.now()
            get_devices_successful = True
        except (OSError, RuntimeError, OSOEnergyApiError, ConnectionError, HTTPException):
            get_devices_successful = False

        return get_devices_successful

    async def start_session(self, config: dict = {}) -> dict[str, list[OSOEnergyWaterHeaterData | OSOEnergySensorData | OSOEnergyBinarySensorData]]:
        # pylint: disable=unused-variable
        """Start session to the OSO Energy platform.

        Args:
            config (dict, optional): Configuration for Home Assistant to use. Defaults to {}.

        Raises:
            OSOEnergyUnknownConfiguration: Unknown configuration identifed.
            OSOEnergyReauthRequired: Subscription key has expired and a new one is required.

        Returns:
            list: List of devices
        """
        custom_component = False
        for file, line, function_name, text in traceback.extract_stack():
            if "/custom_components/" in file:
                custom_component = True

        self.config.sensors = custom_component
        await self.update_interval(30)

        if config != {}:
            if config["api_key"] is not None and not self.config.file:
                await self.update_subscription_key(config["api_key"])
            elif not self.config.file:
                raise OSOEnergyUnknownConfiguration

        try:
            await self.get_devices()
        except HTTPException:
            return HTTPException

        if self.data.devices == {}:
            raise OSOEnergyReauthRequired

        return await self.create_devices()

    async def create_devices(self) -> dict[str, list[OSOEnergyWaterHeaterData | OSOEnergySensorData | OSOEnergyBinarySensorData]]:
        """Create list of devices.

        Returns:
            list: List of devices
        """
        self.device_list["binary_sensor"] = []
        self.device_list["sensor"] = []
        self.device_list["water_heater"] = []
        self.device_list["switch"] = []

        for a_device in self.data["devices"]:
            device = self.data.devices[a_device]
            self.add_device("water_heater", device)

            self.add_sensor("sensor", device, haName=" Heater Mode", osoEnergyType="HEATER_MODE")
            self.add_sensor("sensor", device, haName=" Optimization Mode", osoEnergyType="OPTIMIZATION_MODE")
            self.add_sensor("sensor", device, haName=" Profile", osoEnergyType="PROFILE")
            self.add_sensor("sensor", device, haName=" Volume", osoEnergyType="VOLUME")
            self.add_sensor("sensor", device, haName=" Power Load", osoEnergyType="POWER_LOAD")
            self.add_sensor(
                "sensor",
                device,
                haName=" Tapping Capacity",
                osoEnergyType="TAPPING_CAPACITY"
            )
            self.add_sensor(
                "sensor",
                device,
                haName=" Capacity Mixed Water 40",
                osoEnergyType="CAPACITY_MIXED_WATER_40"
            )
            self.add_sensor("sensor", device, haName=" V40 Min", osoEnergyType="V40_MIN")
            self.add_sensor("sensor", device, haName=" V40 Level Min", osoEnergyType="V40_LEVEL_MIN")
            self.add_sensor("sensor", device, haName=" V40 Level Max", osoEnergyType="V40_LEVEL_MAX")
            if(device.get("control", {}).get("currentTemperatureOne") is not None):
                self.add_sensor("sensor", device, haName=" Temperature One", osoEnergyType="TEMPERATURE_ONE")
            if(device.get("control", {}).get("currentTemperatureLow") is not None):
                self.add_sensor("sensor", device, haName=" Temperature Low", osoEnergyType="TEMPERATURE_LOW")
            if(device.get("control", {}).get("currentTemperatureMid") is not None):
                self.add_sensor("sensor", device, haName=" Temperature Mid", osoEnergyType="TEMPERATURE_MID")
            if(device.get("control", {}).get("currentTemperatureTop") is not None):
                self.add_sensor("sensor", device, haName=" Temperature Top", osoEnergyType="TEMPERATURE_TOP")

            self.add_binary_sensor("binary_sensor", device, haName=" Power Save", osoEnergyType="POWER_SAVE")
            self.add_binary_sensor("binary_sensor", device, haName=" Extra Energy", osoEnergyType="EXTRA_ENERGY")
            self.add_binary_sensor("binary_sensor", device, haName=" Heater State", osoEnergyType="HEATER_STATE")

            self.add_switch("switch", device, haName=" Holiday Mode", osoEnergyType="HOLIDAY_MODE")

        return self.device_list

    def add_device(self, entity_type: str, data: dict):
        """Add entity to the list.

        Args:
            entity_type (str): Type of entity
            data (dict): Information to create entity.
        """
        result = OSOEnergyWaterHeaterData()
        display_name = data.get("deviceName", "Water Heater")
        connection_status = data.get("connectionState", {}).get("connectionState", "Unknown")
        online = OSOTOHA["Hotwater"]["HeaterConnection"].get(connection_status, False)

        try:
            result.ha_name = display_name
            result.ha_type = entity_type
            result.device_id = data["deviceId"]
            result.device_type = data.get("deviceType", "Unknown")
            result.device_name = display_name
            result.power_load = float(data.get("powerConsumption", 0))
            result.volume = float(data.get("volume", 0))
            result.online = online
            result.isInPowerSave = data.get("isInPowerSave", False)
        except KeyError as exception:
            self.logger.error(exception)

        self.device_list[entity_type].append(result)

    def add_sensor(self, entity_type: str, data: dict, haName: str, osoEnergyType: str):
        """Add entity to the list.

        Args:
            entity_type (str): Type of entity
            data (dict): Information to create entity.
            haName (str): Sensor name for HA

        """
        result = OSOEnergySensorData()
        display_name = data.get("deviceName", "Water Heater")
        connection_status = data.get("connectionState", {}).get("connectionState", "Unknown")
        online = OSOTOHA["Hotwater"]["HeaterConnection"].get(connection_status, False)

        try:
            result.ha_name = display_name + haName
            result.ha_type = entity_type
            result.device_id = data["deviceId"]
            result.device_type = data.get("deviceType", "Unknown")
            result.device_name = display_name
            result.online = online
            result.osoEnergyType = osoEnergyType

        except KeyError as exception:
            self.logger.error(exception)

        self.device_list[entity_type].append(result)

    def add_binary_sensor(self, entity_type: str, data: dict, haName: str, osoEnergyType: str):
        """Add entity to the list.

        Args:
            entity_type (str): Type of entity
            data (dict): Information to create entity.
            haName (str): Sensor name for HA

        """
        result = OSOEnergyBinarySensorData()
        display_name = data.get("deviceName", "Water Heater")
        connection_status = data.get("connectionState", {}).get("connectionState", "Unknown")
        online = OSOTOHA["Hotwater"]["HeaterConnection"].get(connection_status, False)

        try:
            result.ha_name = display_name + haName
            result.ha_type = entity_type
            result.device_id = data["deviceId"]
            result.device_type = data.get("deviceType", "Unknown")
            result.device_name = display_name
            result.online = online
            result.osoEnergyType = osoEnergyType

        except KeyError as exception:
            self.logger.error(exception)

        self.device_list[entity_type].append(result)

    def add_switch(self, entity_type: str, data: dict, haName: str, osoEnergyType: str):
        """Add switch to the list.

        Args:
            entity_type (str): Type of entity
            data (dict): Information to create entity.
            haName (str): Sensor name for HA

        """
        result = OSOEnergySwitchData()
        display_name = data.get("deviceName", "Water Heater")
        connection_status = data.get("connectionState", {}).get("connectionState", "Unknown")
        online = OSOTOHA["Hotwater"]["HeaterConnection"].get(connection_status, False)

        try:
            result.ha_name = display_name + haName
            result.ha_type = entity_type
            result.device_id = data["deviceId"]
            result.device_type = data.get("deviceType", "Unknown")
            result.device_name = display_name
            result.online = online
            result.osoEnergyType = osoEnergyType

        except KeyError as exception:
            self.logger.error(exception)

        self.device_list[entity_type].append(result)

    @staticmethod
    def epochTime(date_time: any, pattern: str, action: str):
        # pylint: disable=invalid-name
        """date/time conversion to epoch.

        Args:
            date_time (any): epoch time or date and time to use.
            pattern (str): Pattern for converting to epoch.
            action (str): Convert from/to.

        Returns:
            any: Converted time.
        """
        if action == "to_epoch":
            pattern = "%d.%m.%Y %H:%M:%S"
            epochtime = int(time.mktime(time.strptime(str(date_time), pattern)))
            return epochtime

        date = datetime.fromtimestamp(int(date_time)).strftime(pattern)
        return date



"""Custom Logging Module."""

import inspect
from datetime import datetime


class Logger:
    """Custom Logging Code."""

    def __init__(self, session=None):
        """Initialise the logger class."""
        self.session = session

    async def error(self, error="UNKNOWN"):
        """Process and unexpected error."""
        self.session.logger.error(
            f"An unexpected error has occurred whilst"
            f" executing {inspect.stack()[1][3]}"
            f" with exception {error.__class__} {error}"
        )

    async def error_check(self, n_id, error_type):
        """Error has occurred."""
        message = None

        if error_type is False:
            message = "Device offline could not update entity - " + n_id
            if n_id not in self.session.config.error_list:
                self.session.logger.warning(message)
                self.session.config.error_list.update({n_id: datetime.now()})
        elif error_type == "Failed":
            message = "ERROR - No data found for device - " + n_id
            if n_id not in self.session.config.error_list:
                self.session.logger.error(message)
                self.session.config.error_list.update({n_id: datetime.now()})



"""Dot notation for dictionary."""


class Map(dict):
    """dot.notation access to dictionary attributes.

    Args:
        dict (dict): dictionary to map.
    """

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__



"""Start OSO Energy Session."""

import sys
import traceback
from os.path import expanduser
from typing import Optional

from aiohttp import ClientSession
from loguru import logger
from apyosoenergyapi.sensor import Sensor
from apyosoenergyapi.binary_sensor import BinarySensor
from apyosoenergyapi.switch import Switch

from .session import OSOEnergySession
from .waterheater import WaterHeater
from .device_attributes import OSOEnergyAttributes

debug = []
home = expanduser("~")
logger.add(
    home + "/pyosoenergyapi_debug.log", filter=lambda record: record["level"].name == "DEBUG"
)
logger.add(
    home + "/pyosoenergyapi_info.log", filter=lambda record: record["level"].name == "INFO"
)
logger.add(
    home + "/pyosoenergyapi_error.log", filter=lambda record: record["level"].name == "ERROR"
)


def exception_handler(exctype, value, tb):
    # pylint: disable=unused-argument
    # pylint: disable=invalid-name
    """Handle custom exceptions.

    Args:
        exctype ([type]): [description]
        value ([type]): [description]
        tb ([type]): [description]
    """
    last = len(traceback.extract_tb(tb)) - 1
    logger.error(
        f"-> \n"
        f"Error in {traceback.extract_tb(tb)[last].filename}\n"
        f"when running {traceback.extract_tb(tb)[last].name} function\n"
        f"on line {traceback.extract_tb(tb)[last].lineno} - "
        f"{traceback.extract_tb(tb)[last].line} \n"
        f"with vars {traceback.extract_tb(tb)[last].locals}"
    )
    traceback.print_exc(tb)


sys.excepthook = exception_handler


def trace_debug(frame, event, arg):
    """Trace functions.

    Args:
        frame (object): The current frame being debugged.
        event (str): The event type
        arg (dict): arguments in debug function..
    Returns:
        object: returns itself as per tracing docs
    """
    if "pyosoenergyapi/" in str(frame):
        code = frame.f_code
        func_name = code.co_name
        func_line_no = frame.f_lineno
        if func_name in debug:
            if event == "call":
                func_filename = code.co_filename.rsplit("/", 1)
                caller = frame.f_back
                caller_line_no = caller.f_lineno
                caller_filename = caller.f_code.co_filename.rsplit("/", 1)

                logger.debug(
                    f"Call to {func_name} on line {func_line_no} "
                    f"of {func_filename[1]} from line {caller_line_no} "
                    f"of {caller_filename[1]}"
                )
            elif event == "return":
                logger.debug(f"returning {arg}")

        return trace_debug

    return None


class OSOEnergy(OSOEnergySession):
    """OSO Energy class.

    Args:
        OSOEnergySession (object): Interact with OSO Energy
    """

    def __init__(
            self,
            subscription_key,
            websession: Optional[ClientSession] = None):
        """Initialize OSO Energy."""
        super().__init__(subscription_key=subscription_key, websession=websession)
        self.session = self
        self.attr = OSOEnergyAttributes(self.session)
        self.hotwater = WaterHeater(self.session)
        self.sensor = Sensor(self.session)
        self.binary_sensor = BinarySensor(self.session)
        self.switch = Switch(self.session)
        self.logger = logger
        if debug:
            sys.settrace(trace_debug)

    def setDebugging(self, debugger: list):
        # pylint: disable=no-self-use
        # pylint: disable=global-statement
        # pylint: disable=invalid-name
        """Set function to debug.

        Args:
            debugger (list): a list of functions to debug

        Returns:
            object: Returns traceback object.
        """
        global debug
        debug = debugger
        if debug:
            return sys.settrace(trace_debug)
        return sys.settrace(None)



"""OSO Energy Sensor Module."""

from .helper.const import sensor_commands, OSOEnergySensorData


class OSOEnergySensor:
    # pylint: disable=no-member
    """OSO Energy Sensor Code."""

    sensorType = "Sensor"
    hotwaterType = "Hotwater"
    hotwaterConnection = "HeaterConnection"

class Sensor(OSOEnergySensor):
    """Home Assistant sensor code.

    Args:
        OSOEnergySensor (object): OSO Energy sensor code.
    """

    def __init__(self, session: object = None):
        """Initialise sensor.

        Args:
            session (object, optional): session to interact with OSO Energy. Defaults no None.
        """
        self.session = session

    async def get_sensor(self, device: OSOEnergySensorData) -> OSOEnergySensorData:
        # pylint: disable=eval-used
        """Get updated sensor data.

        Args:
            device (OSOEnergySensorData): Device to update.

        Returns:
            OSOEnergySensorData: Updated device.
        """
        device.online = await self.session.attr.online_offline(device.device_id)
        if device.online:
            self.session.helper.device_recovered(device.device_id)
            dev_data = OSOEnergySensorData()
            dev_data.online = device.online
            dev_data.ha_name = device.ha_name
            dev_data.ha_type = device.ha_type
            dev_data.osoEnergyType = device.osoEnergyType
            dev_data.device_id = device.device_id
            dev_data.device_type = device.device_type
            dev_data.device_name = device.device_name
            dev_data.available = await self.session.attr.online_offline(device.device_id)

            if dev_data.osoEnergyType in sensor_commands:
                code = sensor_commands.get(dev_data.osoEnergyType)
                dev_data.state = await eval(code)

            self.session.sensors.update({device.device_id: dev_data})
            return self.session.sensors[device.device_id]

        await self.session.log.error_check(
            device.device_id, device.online
        )
        return device



"""OSO Energy exception class."""


class OSOEnergyApiError(Exception):
    """Api error.

    Args:
        Exception (object): Exception object to invoke
    """


class OSOEnergyReauthRequired(Exception):
    """Re-Authentication is required.

    Args:
        Exception (object): Exception object to invoke
    """


class OSOEnergyUnknownConfiguration(Exception):
    """Unknown OSO Energy Configuration.

    Args:
        Exception (object): Exception object to invoke
    """


class NoSubscriptionKey(Exception):
    """No Subscription key exception.

    Args:
        Exception (object): Exception object to invoke
    """



"""OSO Energy Water Heater Module."""

from array import array
from numbers import Number
from aiohttp.web_exceptions import HTTPError
from .helper.const import OSOTOHA, OSOEnergyWaterHeaterData
from datetime import datetime, timezone, timedelta


class OSOWaterHeater:
    # pylint: disable=no-member
    """Water Heater Code.

    Returns:
        object: Water Heater Object
    """

    hotwaterType = "Hotwater"

    async def get_heater_state(self, device: OSOEnergyWaterHeaterData):
        """Get water heater current mode.

        Args:
            device (OSOEnergyWaterHeaterData): Device to get the mode for.

        Returns:
            str: Return mode.
        """
        state = None
        final = None

        try:
            device_data = self.session.data.devices[device.device_id]
            state = device_data["control"]["heater"]
            final = OSOTOHA[self.hotwaterType]["HeaterState"].get(state, state)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def turn_on(self, device: OSOEnergyWaterHeaterData, full_utilization: bool):
        """Turn device on.

        Args:
            device (OSOEnergyWaterHeaterData): Device to turn on.
            full_utilization (bool): Fully utilize device.

        Returns:
            boolean: return True/False if turn on was successful.
        """
        final = False

        try:
            resp = await self.session.api.turn_on(device.device_id, full_utilization)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final

    async def turn_off(self, device: OSOEnergyWaterHeaterData, full_utilization: bool):
        """Turn device off.

        Args:
            device (OSOEnergyWaterHeaterData): Device to turn off.
            full_utilization (bool): Fully utilize device.

        Returns:
            boolean: return True/False if turn off was successful.
        """
        final = False

        try:
            resp = await self.session.api.turn_off(device.device_id, full_utilization)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final

    async def set_v40_min(self, device: OSOEnergyWaterHeaterData, v40min: float):
        """Set V40 Min levels for device.

        Args:
            device (OSOEnergyWaterHeaterData): Device to turn off.
            v40Min (float): quantity of water at 40°C.

        Returns:
            boolean: return True/False if setting the V40Min was successful.
        """
        final = False

        try:
            resp = await self.session.api.set_v40_min(device.device_id, v40min)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final

    async def set_optimization_mode(self, device: OSOEnergyWaterHeaterData, option: Number, sub_option: Number):
        """Set heater optimization mode.

        Args:
            device (OSOEnergyWaterHeaterData): Device to turn off.
            option (Number): heater optimization option.
            sub_option (Number): heater optimization sub option.

        Returns:
            boolean: return True/False if setting the optimization mode was successful.
        """
        final = False

        try:
            resp = await self.session.api.set_optimization_mode(
                device.device_id,
                optimizationOptions=option,
                optimizationSubOptions=sub_option
            )
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final

    async def set_profile(self, device: OSOEnergyWaterHeaterData, profile: array):
        """Set heater profile.

        Args:
            device (OSOEnergyWaterHeaterData): Device to set profile to.
            profile (array): array of temperatures for 24 hours (UTC).

        Returns:
            boolean: return True/False if setting the profile was successful.
        """
        final = False

        try:
            resp = await self.session.api.set_profile(device.device_id, hours=profile)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final
    
    async def enable_holiday_mode(self, device: OSOEnergyWaterHeaterData, period_days: int = 365):
        """Enable holiday mode for device.

        Args:
            device (OSOEnergyWaterHeaterData): Device to enable holiday mode for.
            period_days (int, optional): Number of days to enable holiday mode for. Defaults to 365.

        Returns:
            boolean: return True/False if enabling holiday mode was successful.
        """
        final = False
        try:
            start_date = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            end_date = (datetime.now(timezone.utc) + timedelta(days=period_days)).strftime("%Y-%m-%dT%H:%M:%S.%fZ")
            resp = await self.session.api.enable_holiday_mode(device.device_id, start_date, end_date)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final
    
    async def disable_holiday_mode(self, device: OSOEnergyWaterHeaterData):
        """Disable holiday mode for device.

        Args:
            device (OSOEnergyWaterHeaterData): Device to disable holiday mode for.

        Returns:
            boolean: return True/False if disabling holiday mode was successful.
        """
        final = False

        try:
            resp = await self.session.api.disable_holiday_mode(device.device_id)
            if resp["original"] == 200:
                final = True
                await self.session.get_devices()

        except HTTPError as exception:
            await self.session.log.error(exception)

        return final


class WaterHeater(OSOWaterHeater):
    """Water heater class.

    Args:
        OSOWaterHeater (object): OSOWaterHeater class.
    """

    def __init__(self, session: object = None):
        """Initialise water heater.

        Args:
            session (object, optional): Session to interact with account. Defaults to None.
        """
        self.session = session

    async def get_water_heater(self, device: OSOEnergyWaterHeaterData) -> OSOEnergyWaterHeaterData:
        """Update water heater device.

        Args:
            device (OSOEnergyWaterHeaterData): device to update.

        Returns:
            OSOEnergyWaterHeaterData: Updated device.
        """
        device.online = await self.session.attr.online_offline(device.device_id)
        if(device.online):
            self.session.helper.device_recovered(device.device_id)
            
            dev_data = OSOEnergyWaterHeaterData()
            dev_data.ha_name = device.ha_name
            dev_data.ha_type = device.ha_type
            dev_data.device_id = device.device_id
            dev_data.device_type = device.device_type
            dev_data.device_name = device.device_name
            dev_data.current_operation = await self.get_heater_state(device)

            attributes = await self.session.attr.state_attributes(device.device_id)

            dev_data.available = attributes.get("available")
            dev_data.optimization_mode = attributes.get("optimization_mode")
            dev_data.heater_state = attributes.get("heater_state")
            dev_data.heater_mode = attributes.get("heater_mode")
            dev_data.current_temperature = attributes.get("current_temperature")
            dev_data.target_temperature = attributes.get("target_temperature")
            dev_data.target_temperature_low = attributes.get("target_temperature_low")
            dev_data.target_temperature_high = attributes.get("target_temperature_high")
            dev_data.min_temperature = attributes.get("min_temperature")
            dev_data.max_temperature = attributes.get("max_temperature")
            dev_data.profile = attributes.get("profile")
            dev_data.power_load = attributes.get("power_load")
            dev_data.volume = attributes.get("volume")
            dev_data.isInPowerSave = attributes.get("isInPowerSave", False)

            self.session.devices.update({device.device_id: dev_data})
            return self.session.devices[device.device_id]

        await self.session.log.error_check(
            device.device_id, device.online
        )
        return device



"""Setup pyosoenergyapi package."""
import os
import re

import unasync
from setuptools import setup


def requirements_from_file(filename="requirements.txt"):
    """Get requirements from file."""
    with open(os.path.join(os.path.dirname(__file__), filename)) as r:
        reqs = r.read().strip().split("\n")
    # Return non empty lines and non comments
    return [r for r in reqs if re.match(r"^\w+", r)]


setup(
    version="1.2.4",
    cmdclass={
        "build_py": unasync.cmdclass_build_py(
            rules=[
                unasync.Rule(
                    "/apyosoenergyapi/",
                    "/pyosoenergyapi/",
                    additional_replacements={
                        "apyosoenergyapi": "pyosoenergyapi",
                    },
                )
            ]
        )
    },
    install_requires=requirements_from_file()
)


"""OSO Energy Device Attribute Module."""
from typing import Any
from .helper.logger import Logger
from .helper.const import OSOTOHA


class OSOEnergyAttributes:  # pylint: disable=too-many-public-methods
    """Devcie Attributes Code."""

    hotwaterType = "Hotwater"
    hotwaterState = "HeaterState"
    hotwaterConnection = "HeaterConnection"
    hotwaterOptimizationMode = "HeaterOptimizationMode"
    hotwaterSubOptimizationMode = "HeaterSubOptimizationMode"

    def __init__(self, session: object = None):
        """Initialise attributes.

        Args:
            session (object, optional): Session to interact with OSO Energy. Defaults to None.
        """
        self.session = session
        self.session.log = Logger(session)
        self.type = "Attribute"

    async def state_attributes(self, device_id: str) -> dict[str, Any]:
        """Get HS State Attributes.

        Args:
            device_id (str): The id of the device

        Returns:
            dict: Set of attributes
        """
        attr = {}

        if device_id in self.session.data.devices:
            attr.update({"available": (await self.online_offline(device_id))})
            attr.update({"power_load": (await self.get_power_consumption(device_id))})
            attr.update({"volume": (await self.get_volume(device_id))})
            attr.update({"tapping_capacity": (await self.get_tapping_capacity(device_id))})
            attr.update({
                "capacity_mixed_water_40": (await self.get_capacity_mixed_water_40(device_id))
            })
            attr.update({"actual_load_kwh": (await self.get_actual_load_kwh(device_id))})
            attr.update({"heater_state": (await self.get_heater_state(device_id))})
            attr.update({"heater_mode": (await self.get_heater_mode(device_id))})
            attr.update({"current_temperature": (await self.get_current_temperature(device_id))})
            attr.update({"target_temperature": (await self.get_target_temperature(device_id))})
            attr.update({
                "target_temperature_low": (await self.get_target_temperature_low(device_id))
            })
            attr.update({
                "target_temperature_high": (await self.get_target_temperature_high(device_id))
            })
            attr.update({"min_temperature": (await self.get_min_temperature(device_id))})
            attr.update({"max_temperature": (await self.get_max_temperature(device_id))})
            attr.update({"optimization_mode": (await self.get_optimization_mode(device_id))})
            attr.update({"v40_min": (await self.get_v40_min(device_id))})
            attr.update({"v40_level_min": (await self.get_v40_level_min(device_id))})
            attr.update({"v40_level_max": (await self.get_v40_level_max(device_id))})
            attr.update({"profile": (await self.get_profile(device_id))})
            attr.update({"isInPowerSave": (await self.get_power_save_bool(device_id))})

        return attr

    async def get_heater_state_bool(self, device_id: str) -> bool:
        """Get state of heating.

        Args:
            device_id (str): The id of the device

        Returns:
            str: The state of the heater.
        """
        state = None
        final = None

        try:
            data = self.session.data.devices[device_id]
            state = data.get("control", {}).get("heater", 0)
            final = OSOTOHA[self.hotwaterType]["HeaterStateBool"].get(state, False)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final
    
    async def get_power_save_bool(self, device_id: str) -> bool:
        """Check if device is in Power Save mode.

        Args:
            device_id (str): The id of the device.

        Returns:
            boolean: True/False if device in Power Save mode.
        """
        state = None
        final = False

        try:
            data = self.session.data.devices[device_id]
            state = data.get("isInPowerSave", False)
            final = OSOTOHA[self.hotwaterType]["HeaterPowerSaveModeBool"].get(state, False)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def get_extra_energy_bool(self, device_id: str) -> bool:
        """Check if device is in Extra Energy mode.

        Args:
            device_id (str): The id of the device.

        Returns:
            boolean: True/False if device in Extra Energy mode.
        """
        state = None
        final = False

        try:
            data = self.session.data.devices[device_id]
            state = data.get("isInExtraEnergy", False)
            final = OSOTOHA[self.hotwaterType]["HeaterExtraEnergyModeBool"].get(state, False)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def get_power_save(self, device_id: str):
        """Check if device is in Power Save mode.

        Args:
            device_id (str): The id of the device.

        Returns:
            boolean: True/False if device in Power Save mode.
        """
        state = None
        final = False

        try:
            data = self.session.data.devices[device_id]
            state = data.get("isInPowerSave", False)
            final = OSOTOHA[self.hotwaterType]["HeaterPowerSaveMode"].get(state, False)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def get_extra_energy(self, device_id: str):
        """Check if device is in Extra Energy mode.

        Args:
            device_id (str): The id of the device.

        Returns:
            boolean: True/False if device in Extra Energy mode.
        """
        state = None
        final = False

        try:
            data = self.session.data.devices[device_id]
            state = data.get("isInExtraEnergy", False)
            final = OSOTOHA[self.hotwaterType]["HeaterExtraEnergyMode"].get(state, False)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def online_offline(self, device_id: str):
        """Check if device is online.

        Args:
            device_id (str): The id of the device.

        Returns:
            boolean: True/False if device online.
        """
        state = None
        final = False

        try:
            data = self.session.data.devices[device_id]
            state = data["connectionState"]["connectionState"]
            final = OSOTOHA[self.hotwaterType][self.hotwaterConnection].get(state, False)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def get_power_consumption(self, device_id: str):
        """Get heater power consumption.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The mode of the device.
        """
        consumption = None

        try:
            data = self.session.data.devices[device_id]
            consumption = data.get("powerConsumption", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return consumption

    async def get_volume(self, device_id: str):
        """Get heater volume.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The volume of the device.
        """
        volume = None

        try:
            data = self.session.data.devices[device_id]
            volume = data.get("volume", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return volume

    async def get_tapping_capacity(self, device_id: str):
        """Get tapping capacity in kWh.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The tapping capacity kWh.
        """
        capacity = None

        try:
            data = self.session.data.devices[device_id]
            capacity = data.get("data", {}).get("tappingCapacitykWh", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return capacity

    async def get_capacity_mixed_water_40(self, device_id: str):
        """Get capacity of water at 40 degrees.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The capacity of water at 40 degrees.
        """
        capacity = None

        try:
            data = self.session.data.devices[device_id]
            capacity = data.get("data", {}).get("capacityMixedWater40", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return capacity

    async def get_actual_load_kwh(self, device_id: str):
        """Get load of heater in kW.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The actual load of the heater.
        """
        load = None

        try:
            data = self.session.data.devices[device_id]
            load = data.get("data", {}).get("actualLoadKwh", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return load

    async def get_heater_state(self, device_id: str):
        """Get state of heating.

        Args:
            device_id (str): The id of the device

        Returns:
            str: The state of the heater.
        """
        state = None
        final = None

        try:
            data = self.session.data.devices[device_id]
            state = data.get("control", {}).get("heater", 0)
            final = OSOTOHA[self.hotwaterType][self.hotwaterState].get(state, "OFF")
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def get_heater_mode(self, device_id: str):
        """Get mode of heater.

        Args:
            device_id (str): The id of the device

        Returns:
            str: The mode of the heater.
        """
        state = None
        final = None

        try:
            data = self.session.data.devices[device_id]
            state = data.get("control", {}).get("mode", None)
            final = OSOTOHA[self.hotwaterType]["HeaterMode"].get(state, state)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def get_current_temperature(self, device_id: str):
        """Get current temperature of heater.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The current temperature of the heater.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("currentTemperature", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature

    async def get_target_temperature(self, device_id: str):
        """Get current target temperature of heater.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The current target temperature of the heater.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("targetTemperature", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature

    async def get_target_temperature_low(self, device_id: str):
        """Get current target temperature low of heater.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The current target temperature low of the heater.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("targetTemperatureLow", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature

    async def get_target_temperature_high(self, device_id: str):
        """Get current target temperature high of heater.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The current target temperature high of the heater.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("targetTemperatureHigh", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature

    async def get_min_temperature(self, device_id: str):
        """Get min temperature of heater.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The min temperature of the heater.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("minTemperature", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature

    async def get_max_temperature(self, device_id: str):
        """Get max temperature of heater.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The max temperature of the heater.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("maxTemperature", 0)
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature

    async def get_optimization_mode(self, device_id: str):
        """Get heater optimization mode.

        Args:
            device_id (str): The id of the device

        Returns:
            str: The optimization mode of the device.
        """
        mode = None
        final = None

        try:
            data = self.session.data.devices[device_id]
            mode = data["optimizationOption"]
            final = OSOTOHA[self.hotwaterType][self.hotwaterOptimizationMode].get(mode, mode)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def get_sub_optimization_mode(self, device_id: str):
        """Get heater sub optimization mode.

        Args:
            device_id (str): The id of the device

        Returns:
            str: The sub optimization mode of the device.
        """
        mode = None
        final = None

        try:
            data = self.session.data.devices[device_id]
            mode = data["optimizationSubOption"]
            final = OSOTOHA[self.hotwaterType][self.hotwaterSubOptimizationMode].get(mode, mode)
        except KeyError as exception:
            await self.session.log.error(exception)

        return final

    async def get_v40_min(self, device_id: str):
        """Get v40 min level of the heater.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The v40 min level of the device.
        """
        level = None

        try:
            data = self.session.data.devices[device_id]
            level = data["v40Min"]
        except KeyError as exception:
            await self.session.log.error(exception)

        return level

    async def get_v40_level_min(self, device_id: str):
        """Get v40 level min.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The minimum possible level of v40 of the device.
        """
        level = None

        try:
            data = self.session.data.devices[device_id]
            level = data["v40LevelMin"]
        except KeyError as exception:
            await self.session.log.error(exception)

        return level

    async def get_v40_level_max(self, device_id: str):
        """Get v40 level max.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The maximum possible level of v40 of the device.
        """
        level = None

        try:
            data = self.session.data.devices[device_id]
            level = data["v40LevelMax"]
        except KeyError as exception:
            await self.session.log.error(exception)

        return level

    async def get_profile(self, device_id: str):
        """Get the 24 hour profile of the heater (UTC).

        Args:
            device_id (str): The id of the device

        Returns:
            float array: The 24 hour temperature profile of the device. (UTC)
        """
        level = None

        try:
            data = self.session.data.devices[device_id]
            level = data["profile"]
        except KeyError as exception:
            await self.session.log.error(exception)

        return level

    async def get_temperature_one(self, device_id: str):
        """Get the one temperature.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The current reported temperature from the one wire sensor.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("currentTemperatureOne")
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature
    
    async def get_temperature_low(self, device_id: str):
        """Get the low temperature.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The current reported temperature from the low sensor.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("currentTemperatureLow")
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature
    
    async def get_temperature_mid(self, device_id: str):
        """Get the mid temperature.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The current reported temperature from the mid sensor.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("currentTemperatureMid")
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature
    
    async def get_temperature_top(self, device_id: str):
        """Get the top temperature.

        Args:
            device_id (str): The id of the device

        Returns:
            float: The current reported temperature from the top sensor.
        """
        temperature = None

        try:
            data = self.session.data.devices[device_id]
            temperature = data.get("control", {}).get("currentTemperatureTop")
        except KeyError as exception:
            await self.session.log.error(exception)

        return temperature


"""OSO Energy Helper code."""


class OSOEnergyHelper:  # pylint: disable=too-few-public-methods
    """OSO Energy helper class."""

    def __init__(self, session: object = None):
        """OSO Energy Helper.

        Args:
            session (object, optional): Interact with OSO Energy. Defaults to None.
        """
        self.session = session

    def device_recovered(self, n_id: str):
        """Register that device has recovered from being offline.

        Args:
            n_id (str): ID of the device
        """
        if n_id in self.session.config.error_list:
            self.session.config.error_list.pop(n_id)
